import * as METADATA from '../constants';
import { DecoratorType, POUCH_DEFAULT_QUALIFIER } from '../constants';
import InjectConfig from '../decorator/InjectConfig';

export interface InjectItem {
  decoratorType: DecoratorType;
  propertyConfig?: InjectConfig;
  parameterConfigMap?: Map<number, InjectConfig>;
}

class PouchInjectDefinition {
  protected target: Object;
  protected injectMap: Map<string | symbol, InjectItem>;

  public constructor(target: Object) {
    this.target = target;
    this.injectMap = new Map();
  }

  public get injectors(): Map<string | symbol, InjectItem> {
    return this.injectMap;
  }

  /**
   * addInject
   */
  public addInject(injectConfig: InjectConfig, propertyKey: string | symbol, parameterIndex?: number): void {
    if (undefined === injectConfig.qualifier) {
      injectConfig.qualifier = POUCH_DEFAULT_QUALIFIER;
    }
    if (undefined === injectConfig.required) {
      injectConfig.required = false;
    }

    if(typeof parameterIndex === 'number') {
      this.addInjectParameter(injectConfig, propertyKey, parameterIndex);
    } else {
      this.addInjectProperty(injectConfig, propertyKey);
    }
  }

  private addInjectParameter(injectConfig: InjectConfig, propertyKey: string | symbol, parameterIndex: number): void {
    let injectItem: InjectItem;
    let parameterTypes: any[] = Reflect.getMetadata(METADATA.DESIGN_PARAMTYPES, this.target, propertyKey);

    if (undefined === injectConfig.type) {
      injectConfig.type = parameterTypes[parameterIndex];
    }

    if (this.injectMap.has(propertyKey)) {
      injectItem = this.injectMap.get(propertyKey);
    } else {
      injectItem = {
        decoratorType: DecoratorType.PARAMETER,
      };
      this.injectMap.set(propertyKey, injectItem);
    }

    if (null !== injectItem.parameterConfigMap) {
      if (injectItem.parameterConfigMap.has(parameterIndex)) {
        throw new Error(`Cannot apply @inject decorator on one parameter[${propertyKey}:${parameterIndex}] multiple times.`);
      }
    } else {
      injectItem.parameterConfigMap = new Map();
    }

    injectItem.parameterConfigMap.set(parameterIndex, injectConfig);

  }

  private addInjectProperty(injectConfig: InjectConfig, propertyKey: string | symbol): void {
    if (this.injectMap.has(propertyKey)) {
      throw new Error(`Cannot apply @inject decorator on one property[${propertyKey}] multiple times.`);
    }

    let propertyType: any = Reflect.getMetadata(METADATA.DESIGN_TYPE, this.target, propertyKey);

    if (undefined === injectConfig.type) {
      injectConfig.type = propertyType;
    }

    let injectItem: InjectItem = {
      decoratorType: DecoratorType.PROPERTY,
      propertyConfig: injectConfig,
    };

    this.injectMap.set(propertyKey, injectItem);
  }

}

export default PouchInjectDefinition;
